import errno
import ipaddress

import pytest

import doubles
import ruddr
from ruddr import PublishError, FatalPublishError


@pytest.fixture
def mock_addrfile_factory(mocker, tmp_path):
    """Fixture that sets up and returns a mock Addrfile factory"""
    count = 0

    # Arguments set up the initial addresses. Should be of the form:
    # - <name>_set4=ipaddress.IPv4Address(...)|None
    # - <name>_set6=ipaddress.IPv6Network(...)|None
    # - <name>_inv4=ipaddress.IPv4Address(...)|None
    # - <name>_inv6=ipaddress.IPv6Network(...)|None
    def factory(**kwargs):
        nonlocal count
        count += 1
        wrapped_addrfile = ruddr.Addrfile(tmp_path / f"addrfile_{count}")

        for arg, val in kwargs.items():
            name, _, kind = arg.rpartition('_')
            if kind == 'set4':
                wrapped_addrfile.set_ipv4(name, val)
            elif kind == 'set6':
                wrapped_addrfile.set_ipv6(name, val)
            elif kind == 'inv4':
                wrapped_addrfile.invalidate_ipv4(name, val)
            elif kind == 'inv6':
                wrapped_addrfile.invalidate_ipv6(name, val)
            else:
                raise TypeError("Addrfile factory args must end in _set4, "
                                "_set6, _inv4, _inv6")

        addrfile_mock = mocker.MagicMock(spec_set=wrapped_addrfile,
                                         wraps=wrapped_addrfile)
        return addrfile_mock

    return factory


def test_initial_update_ipv4(mock_addrfile_factory, advance):
    """Test initial update updates IPv4 if not current and updates addrfile"""
    addrfile = mock_addrfile_factory(
        test_updater_inv4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.initial_update()

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4')]
    assert updater.ipv6s_published == []
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == []
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []


def test_initial_update_ipv6(mock_addrfile_factory, advance):
    """Test initial update updates IPv6 if not current and updates addrfile"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_inv6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.initial_update()

    assert advance.count_running() == 0
    assert updater.ipv4s_published == []
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == []
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]


def test_initial_update_not_necessary(mock_addrfile_factory, advance):
    """Test initial update does no updates if current"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.initial_update()

    assert advance.count_running() == 0
    assert updater.ipv4s_published == []
    assert updater.ipv6s_published == []
    assert addrfile.invalidate_ipv4.call_args_list == []
    assert addrfile.invalidate_ipv6.call_args_list == []
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_initial_update_retries_ipv4(mock_addrfile_factory, advance):
    """Test initial update retries IPv4 on failure, invalidating the address in
    the addrfile, even when IPv6 succeeds"""
    addrfile = mock_addrfile_factory(
        test_updater_inv4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_inv6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile,
                                  ipv4_errors=[PublishError, None])
    updater.initial_update()

    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]

    advance.by(300)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4'),
                                       ipaddress.IPv4Address('1.2.3.4')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]


def test_initial_update_retries_ipv6(mock_addrfile_factory, advance):
    """Test initial update retries IPv6 on failure, invalidating the address in
    the addrfile, even when IPv4 succeeds"""
    addrfile = mock_addrfile_factory(
        test_updater_inv4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_inv6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile,
                                  ipv6_errors=[PublishError, None])
    updater.initial_update()

    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64'),
                                       ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]


def test_initial_update_no_retry_after_fatal_ipv4(mock_addrfile_factory,
                                                  advance):
    """Test initial update stops retrying both address families after a fatal
    error in IPv4"""
    addrfile = mock_addrfile_factory(
        test_updater_inv4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_inv6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile,
                                  ipv4_errors=[FatalPublishError, None],
                                  ipv6_errors=[PublishError, None])
    updater.initial_update()

    # Don't assert IPv6 publish attempt was made for certain because it may not
    # happen until after IPv4 fatal error
    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4')]
    assert len(updater.ipv6s_published) <= 1
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_count <= 1
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('1.2.3.4')]
    assert len(updater.ipv6s_published) <= 1
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('1.2.3.4')),),
    ]
    assert addrfile.invalidate_ipv6.call_count <= 1
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_initial_update_no_retry_after_fatal_ipv6(mock_addrfile_factory,
                                                  advance):
    """Test initial update stops retrying both address families after a fatal
    error in IPv6"""
    addrfile = mock_addrfile_factory(
        test_updater_inv4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_inv6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile,
                                  ipv4_errors=[PublishError, None],
                                  ipv6_errors=[FatalPublishError, None])
    updater.initial_update()

    # Don't assert IPv4 publish attempt was made for certain because it may not
    # happen until after IPv6 fatal error
    assert len(updater.ipv4s_published) <= 1
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_count <= 1
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert advance.count_running() == 0
    assert len(updater.ipv4s_published) <= 1
    assert updater.ipv6s_published == [ipaddress.IPv6Network('1234::/64')]
    assert addrfile.invalidate_ipv4.call_count <= 1
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('1234::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv4_succeeds(mock_addrfile_factory, advance):
    """Test update_ipv4 publishes address if not current and invalidates and
    updates addrfile"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]


def test_update_ipv6_succeeds(mock_addrfile_factory, advance):
    """Test update_ipv6 publishes address if not current and invalidates and
    updates addrfile"""
    addrfile = mock_addrfile_factory(
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert advance.count_running() == 0
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]


def test_update_ipv4_same_addr(mock_addrfile_factory, advance):
    """Test update_ipv4 does not publish same address that's already
    published"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.update_ipv4(ipaddress.IPv4Address('1.2.3.4'))

    assert advance.count_running() == 0
    assert updater.ipv4s_published == []
    assert addrfile.invalidate_ipv4.call_args_list == []
    assert addrfile.set_ipv4.call_args_list == []


def test_update_ipv6_same_addr(mock_addrfile_factory, advance):
    """Test update_ipv6 does not publish same address that's already
    published"""
    addrfile = mock_addrfile_factory(
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater('test_updater', addrfile)
    updater.update_ipv6(ipaddress.IPv6Network('1234::/64'))

    assert advance.count_running() == 0
    assert updater.ipv6s_published == []
    assert addrfile.invalidate_ipv6.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv4_not_implemented(mock_addrfile_factory, advance):
    """Test update_ipv4 does not set address in addrfile when not
    implemented"""
    addrfile = mock_addrfile_factory()
    updater = doubles.MockUpdater('test_updater', addrfile,
                                  ipv4_implemented=False)
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.set_ipv4.call_args_list == []


def test_update_ipv6_not_implemented(mock_addrfile_factory, advance):
    """Test update_ipv6 does not set address in addrfile when not
    implemented"""
    addrfile = mock_addrfile_factory()
    updater = doubles.MockUpdater('test_updater', addrfile,
                                  ipv6_implemented=False)
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert advance.count_running() == 0
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv4_retries(mock_addrfile_factory, advance):
    """Test update_ipv4 retries after failure, even when IPv6 succeeds, and
    invalidates old address until succeeding, at which point it stops retrying
    and updates the addrfile"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, None]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]

    advance.by(600)
    assert advance.count_running() == 0

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]


def test_update_ipv6_retries(mock_addrfile_factory, advance):
    """Test update_ipv6 retries after failure, even when IPv4 succeeds, and
    invalidates old address until succeeding, at which point it stops retrying
    and updates the addrfile"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv6_errors=[PublishError, PublishError, None]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
    ]

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
    ]

    advance.by(600)
    assert advance.count_running() == 0

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]


def test_update_ipv4_retries_after_same(mock_addrfile_factory, advance):
    """Test update_ipv4 retries after failure at the originally scheduled time
    after another update for the same address"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, None]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []

    advance.by(100)
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []

    advance.by(200)
    assert advance.count_running() == 0

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]


def test_update_ipv6_retries_after_same(mock_addrfile_factory, advance):
    """Test update_ipv6 retries after failure at the originally scheduled time
    after another update for the same address"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv6_errors=[PublishError, None]
    )
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(100)
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(200)
    assert advance.count_running() == 0

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]


def test_update_ipv4_retry_resets_after_new(mock_addrfile_factory, advance):
    """Test update_ipv4 retry interval returns to minimum after a new failed
    call"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, PublishError, PublishError,
                     None]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []

    advance.by(600)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []

    advance.by(1100)
    updater.update_ipv4(ipaddress.IPv4Address('6.7.8.9'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('6.7.8.9')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('6.7.8.9')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []

    advance.by(300)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('6.7.8.9'),
                                       ipaddress.IPv4Address('6.7.8.9')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('6.7.8.9')),),
        (('test_updater', ipaddress.IPv4Address('6.7.8.9')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('6.7.8.9')),),
    ]


def test_update_ipv6_retry_resets_after_new(mock_addrfile_factory, advance):
    """Test update_ipv6 retry interval returns to minimum after a new failed
    call"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv6_errors=[PublishError, PublishError, PublishError, PublishError,
                     None]
    )
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(600)

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(1100)
    updater.update_ipv6(ipaddress.IPv6Network('6789::/64'))

    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('6789::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('6789::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert advance.count_running() == 0
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('6789::/64'),
                                       ipaddress.IPv6Network('6789::/64')]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('6789::/64')),),
        (('test_updater', ipaddress.IPv6Network('6789::/64')),),
    ]
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('6789::/64')),),
    ]


def test_update_ipv4_retry_stops_after_fatal(mock_addrfile_factory, advance):
    """Test update_ipv4 stops retrying (and update_ipv6 as well) after a
    FatalPublishError"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, FatalPublishError],
        ipv6_errors=[PublishError, PublishError, PublishError]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    advance.by(1)
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(600)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv6_retry_stops_after_fatal(mock_addrfile_factory, advance):
    """Test update_ipv6 stops retrying (and update_ipv4 as well) after a
    FatalPublishError"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, PublishError],
        ipv6_errors=[PublishError, PublishError, FatalPublishError]
    )
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))
    advance.by(1)
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(600)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv4_addrfile_write_error_1(mock_addrfile_factory, advance):
    """Test update_ipv4 stops retrying (and update_ipv6 as well) after a write
    error in the addrfile for .invalidate_ipv4"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, PublishError],
        ipv6_errors=[PublishError, PublishError, PublishError]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    advance.by(1)
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    addrfile.invalidate_ipv4.side_effect = OSError(errno.ETIMEDOUT, "timeout")
    advance.by(600)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv6_addrfile_write_error_1(mock_addrfile_factory, advance):
    """Test update_ipv6 stops retrying (and update_ipv4 as well) after a write
    error in the addrfile for .invalidate_ipv6"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, PublishError],
        ipv6_errors=[PublishError, PublishError, PublishError]
    )
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))
    advance.by(1)
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    addrfile.invalidate_ipv6.side_effect = OSError(errno.ETIMEDOUT, "timeout")
    advance.by(600)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv4_addrfile_write_error_2(mock_addrfile_factory, advance):
    """Test update_ipv4 stops retrying (and update_ipv6 as well) after a write
    error in the addrfile for .set_ipv4"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, None],
        ipv6_errors=[PublishError, PublishError, PublishError]
    )
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))
    advance.by(1)
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    addrfile.set_ipv4.side_effect = OSError(errno.ETIMEDOUT, "timeout")
    advance.by(600)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.set_ipv6.call_args_list == []


def test_update_ipv6_addrfile_write_error_2(mock_addrfile_factory, advance):
    """Test update_ipv6 stops retrying (and update_ipv4 as well) after a write
    error in the addrfile for .set_ipv6"""
    addrfile = mock_addrfile_factory(
        test_updater_set4=ipaddress.IPv4Address('1.2.3.4'),
        test_updater_set6=ipaddress.IPv6Network('1234::/64'),
    )
    updater = doubles.MockUpdater(
        'test_updater', addrfile,
        ipv4_errors=[PublishError, PublishError, PublishError],
        ipv6_errors=[PublishError, PublishError, None]
    )
    updater.update_ipv6(ipaddress.IPv6Network('5678::/64'))
    advance.by(1)
    updater.update_ipv4(ipaddress.IPv4Address('5.6.7.8'))

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    advance.by(300)

    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == []

    addrfile.set_ipv6.side_effect = OSError(errno.ETIMEDOUT, "timeout")
    advance.by(600)

    assert advance.count_running() == 0
    assert updater.ipv4s_published == [ipaddress.IPv4Address('5.6.7.8'),
                                       ipaddress.IPv4Address('5.6.7.8')]
    assert updater.ipv6s_published == [ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64'),
                                       ipaddress.IPv6Network('5678::/64')]
    assert addrfile.invalidate_ipv4.call_args_list == [
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
        (('test_updater', ipaddress.IPv4Address('5.6.7.8')),),
    ]
    assert addrfile.invalidate_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]
    assert addrfile.set_ipv4.call_args_list == []
    assert addrfile.set_ipv6.call_args_list == [
        (('test_updater', ipaddress.IPv6Network('5678::/64')),),
    ]



"""Helper function for notifiers to look up IP addresses for the current
system's interfaces"""

# Netifaces is no longer currently maintained (looking for a new maintainer).
# But the last release was in 2021 and it's not that complex, especially the
# address lookup part, which hasn't been modified in several releases.
#
# If this gets to be a problem, a possible alternative is using ctypes or
# similar to call getifaddrs(3) directly, but that's basically what netifaces
# does already, in native C (and actually, it also has alternative
# implementations for Windows, AIX, and Solaris, which don't have getifaddrs).
#
# Some more information on this at:
# https://stackoverflow.com/questions/20743709/get-ipv6-addresses-in-linux-using-ioctl

import ipaddress
from typing import Tuple, List, cast, Dict

import netifaces


def _get_iface_addrs(
    if_name: str
) -> Tuple[List[ipaddress.IPv4Address], List[ipaddress.IPv6Address]]:
    """Lookup current addresses for the named interface.

    :param if_name: Name of the address to look up
    :return: A 2-tuple containing a list of IPV4Address followed by a list
             of IPV6Address.
    :raises ValueError: if there is no interface with the given name.
    """
    # Cast due to lack of type stubs for netifaces
    addresses = cast(Dict[int, List[Dict[str, str]]],
                     netifaces.ifaddresses(if_name))
    ipv4 = [a['addr'] for a in addresses[netifaces.AF_INET]]
    ipv6 = [a['addr'] for a in addresses[netifaces.AF_INET6]]

    # ipaddress.IPv6Address gets upset when there is %ifacename at the end
    # of an address in Python < 3.9. Chop it off.
    ipv6 = [a.partition('%')[0] for a in ipv6]

    ipv4 = [ipaddress.IPv4Address(a) for a in ipv4]
    ipv6 = [ipaddress.IPv6Address(a) for a in ipv6]

    return (ipv4, ipv6)


def get_iface_addrs(
    if_name: str,
    omit_private: bool = True,
    omit_link_local: bool = True
) -> Tuple[List[ipaddress.IPv4Address], List[ipaddress.IPv6Address]]:
    """Lookup current addresses for the named interface. Results will be
    ordered with non-private addresses first, then private (if requested), then
    link-local (if requested).

    :param if_name: Name of the address to look up
    :param omit_private: Whether to omit private addresses (including link- and
                         site-local)
    :param omit_link_local: Whether to omit link-local addresses
    :return: A 2-tuple containing a list of IPV4Address followed by a list
             of IPV6Address.
    :raises ValueError: if there is no interface with the given name.
    """
    ipv4, ipv6 = _get_iface_addrs(if_name)

    result_ipv4 = []
    result_ipv6 = []
    ipv4_private = []
    ipv6_private = []
    ipv4_link_local = []
    ipv6_link_local = []

    for a in ipv4:
        if a.is_link_local:
            ipv4_link_local.append(a)
        elif a.is_private:
            ipv4_private.append(a)
        else:
            result_ipv4.append(a)
    for a in ipv6:
        if a.is_link_local:
            ipv6_link_local.append(a)
        elif a.is_private:
            ipv6_private.append(a)
        else:
            result_ipv6.append(a)

    if not omit_private:
        result_ipv4 += ipv4_private
        result_ipv6 += ipv6_private
    if not omit_link_local:
        result_ipv4 += ipv4_link_local
        result_ipv6 += ipv6_link_local

    return (ipv4, ipv6)



"""Ruddr, the Robotic Updater for Dynamic DNS Records

Top-level module, containing classes and objects useful to custom notifiers and
updaters.
"""

from .addrfile import Addrfile
from .configuration import Config, read_config, read_config_from_path
from .exceptions import (RuddrException, RuddrSetupError, ConfigError,
                         NotifierSetupError, NotStartedError, NotifyError,
                         PublishError, FatalPublishError)
from .manager import DDNSManager
from .notifiers import BaseNotifier, Notifier
from .updaters import (Updater, BaseUpdater, OneWayUpdater, TwoWayUpdater,
                       TwoWayZoneUpdater)



import ipaddress
from typing import Dict

import doubles
import ruddr.manager

import pytest


class TestUpdaterNotifierValidation:
    @pytest.mark.parametrize('expected', [True, False])
    def test_validate_notifier_type(self, mocker, notifier_factory, expected):
        """Test validate_notifier_type calls _validate_updater_or_notifier_type
        with 'notifier'"""
        validator = mocker.patch(
            "ruddr.manager._validate_updater_or_notifier_type",
            return_value=expected,
        )
        notifiers = {
            'test': notifier_factory(),
        }
        mocker.patch("ruddr.manager.notifiers.notifiers", new=notifiers)

        result = ruddr.manager.validate_notifier_type('testmod', 'testtype')

        assert result == expected
        assert validator.call_args == (
            ("notifier", notifiers, 'testmod', 'testtype'),
        )

    @pytest.mark.parametrize('expected', [True, False])
    def test_validate_updater_type(self, mocker, updater_factory, expected):
        """Test validate_updater_type calls _validate_updater_or_notifier_type
        with 'updater'"""
        validator = mocker.patch(
            "ruddr.manager._validate_updater_or_notifier_type",
            return_value=expected,
        )
        updaters = {
            'test': updater_factory(),
        }
        mocker.patch("ruddr.manager.updaters.updaters", new=updaters)

        result = ruddr.manager.validate_updater_type('testmod', 'testtype')

        assert result == expected
        assert validator.call_args == (
            ("updater", updaters, 'testmod', 'testtype'),
        )

    def test_built_in(self):
        """Test _validate_updater_or_notifier_type with a "built-in"
        notifier"""
        notifiers = {
            'test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, None, 'test'
        )
        assert result
        assert notifiers == {
            'test': doubles.FakeNotifier,
        }

    def test_entry_point(self):
        """Test _validate_updater_or_notifier_type with an entry_point
        notifier"""
        notifiers = {
            'test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, None, '_test'
        )
        assert result
        assert notifiers == {
            'test': doubles.FakeNotifier,
            '_test': ruddr.notifiers.static.StaticNotifier,
        }

    def test_built_in_entry_point_conflict(self):
        """Test _validate_updater_or_notifier_type with a type matching both a
        built-in and an entry_point notifier and ensure built-in isn't
        replaced"""
        notifiers = {
            '_test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, None, '_test'
        )
        assert result
        assert notifiers == {
            '_test': doubles.FakeNotifier,
        }

    def test_no_such_type(self):
        """Test _validate_updater_or_notifier_type with a type that doesn't
        match anything"""
        notifiers = {
            'test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, None, 'test_invalid'
        )
        assert not result
        assert notifiers == {
            'test': doubles.FakeNotifier,
        }

    def test_module_and_class(self):
        """Test _validate_updater_or_notifier_type with a module and class"""
        notifiers = {
            'test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, 'doubles', 'MockNotifier'
        )
        assert result
        assert notifiers == {
            'test': doubles.FakeNotifier,
            ('doubles', 'MockNotifier'): doubles.MockNotifier
        }

    def test_module_not_importable(self):
        """Test _validate_updater_or_notifier_type with a module and class but
        module not importable"""
        notifiers = {
            'test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, 'invalid', 'MockNotifier'
        )
        assert not result
        assert notifiers == {
            'test': doubles.FakeNotifier,
        }

    def test_class_not_in_module(self):
        """Test _validate_updater_or_notifier_type with a module and class but
        class not in module"""
        notifiers = {
            'test': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, 'doubles', 'InvalidNotifier'
        )
        assert not result
        assert notifiers == {
            'test': doubles.FakeNotifier,
        }

    def test_module_and_class_already_imported(self):
        """Test _validate_updater_or_notifier_type with a module and class
        already imported"""
        notifiers = {
            'test': doubles.FakeNotifier,
            ('doubles', 'MockNotifier'): doubles.MockNotifier
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, 'doubles', 'MockNotifier'
        )
        assert result
        assert notifiers == {
            'test': doubles.FakeNotifier,
            ('doubles', 'MockNotifier'): doubles.MockNotifier,
        }

    def test_class_matches_built_in(self):
        """Test _validate_updater_or_notifier_type with a module and class
        where class name matches a built-in notifier"""
        notifiers = {
            'MockNotifier': doubles.FakeNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, 'doubles', 'MockNotifier'
        )
        assert result
        assert notifiers == {
            'MockNotifier': doubles.FakeNotifier,
            ('doubles', 'MockNotifier'): doubles.MockNotifier,
        }

    def test_class_imported_matches_built_in(self):
        """Test _validate_updater_or_notifier_type with a module and class
        already imported where class name matches a built-in notifier"""
        notifiers = {
            'MockNotifier': doubles.FakeNotifier,
            ('doubles', 'MockNotifier'): doubles.MockNotifier,
        }
        result = ruddr.manager._validate_updater_or_notifier_type(
            'notifier', notifiers, 'doubles', 'MockNotifier'
        )
        assert result
        assert notifiers == {
            'MockNotifier': doubles.FakeNotifier,
            ('doubles', 'MockNotifier'): doubles.MockNotifier,
        }


class TestDDNSManager:
    def _new_mock_notifier(self, name: str, *args, **kwargs):
        """Keep track of MockNotifiers created using mock_built_ins"""
        notifier = doubles.MockNotifier(name, *args, **kwargs)
        self.notifiers[name] = notifier
        return notifier

    @pytest.fixture(autouse=True)
    def mock_built_ins(self, mocker):
        """Fixture setting up the lists of built-in notifiers and updaters with
        mocks"""
        self.notifiers: Dict[str, doubles.MockNotifier] = dict()

        notifiers = {
            'test': doubles.FakeNotifier,
            'mock': self._new_mock_notifier,
        }
        updaters = {
            'test': doubles.MockBaseUpdater,
        }

        mocker.patch("ruddr.manager.notifiers.notifiers", new=notifiers)
        mocker.patch("ruddr.manager.updaters.updaters", new=updaters)

    def test_manager_creates_built_in_notifier(self):
        """Test that DDNSManager creates a notifier using type only and passes
        in its name and config"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                    'test_key': 'test_val',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier': 'test_notifier',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        assert len(manager.notifiers) == 1
        assert type(manager.notifiers['test_notifier']) == doubles.FakeNotifier
        assert manager.notifiers['test_notifier'].name == 'test_notifier'
        assert (manager.notifiers['test_notifier'].config['test_key'] ==
                'test_val')

    def test_manager_creates_custom_notifier(self):
        """Test that DDNSManager creates a notifier using module and type and
        passes in its name and config"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'module': 'doubles',
                    'type': 'FakeNotifier',
                    'test_key': 'test_val',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier': 'test_notifier',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        assert len(manager.notifiers) == 1
        assert (type(manager.notifiers['test_notifier']) ==
                doubles.FakeNotifier)
        assert manager.notifiers['test_notifier'].name == 'test_notifier'
        assert (manager.notifiers['test_notifier'].config['test_key'] ==
                'test_val')

    def test_manager_creates_built_in_updater(self):
        """Test that DDNSManager creates an updater using type only and passes
        in its name and config"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier': 'test_notifier',
                    'test_key': 'test_val',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        assert len(manager.updaters) == 1
        assert (type(manager.updaters['test_updater']) ==
                doubles.MockBaseUpdater)
        assert manager.updaters['test_updater'].name == 'test_updater'
        assert (manager.updaters['test_updater'].config['test_key'] ==
                'test_val')
        assert manager.updaters['test_updater'].addrfile == manager.addrfile

    def test_manager_creates_custom_updater(self):
        """Test that DDNSManager creates an updater using module and type and
        passes in its name and config"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                }
            },
            updaters={
                'test_updater': {
                    'module': 'doubles',
                    'type': 'MockBaseUpdater',
                    'notifier': 'test_notifier',
                    'test_key': 'test_val',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        assert len(manager.updaters) == 1
        assert (type(manager.updaters['test_updater']) ==
                doubles.MockBaseUpdater)
        assert manager.updaters['test_updater'].name == 'test_updater'
        assert (manager.updaters['test_updater'].config['test_key'] ==
                'test_val')
        assert manager.updaters['test_updater'].addrfile == manager.addrfile

    def test_manager_attaches_notifier4(self):
        """Test that DDNSManager attaches a notifier to an updater with
        notifier4"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier4': 'test_notifier',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        notifier = manager.notifiers['test_notifier']
        updater = manager.updaters['test_updater']
        notifier.notify_ipv4(ipaddress.IPv4Address('1.2.3.4'))
        notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
        assert updater.published_addresses == [
            ipaddress.IPv4Address('1.2.3.4'),
        ]

    def test_manager_attaches_notifier6(self):
        """Test that DDNSManager attaches a notifier to an updater with
        notifier6"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier6': 'test_notifier',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        notifier = manager.notifiers['test_notifier']
        updater = manager.updaters['test_updater']
        notifier.notify_ipv4(ipaddress.IPv4Address('1.2.3.4'))
        notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
        assert updater.published_addresses == [
            ipaddress.IPv6Network('1234::/64'),
        ]

    def test_manager_attaches_notifier4_and_notifier6(self):
        """Test that DDNSManager attaches a notifier to an updater with
        notifier4 and notifier6"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier4': 'test_notifier',
                    'notifier6': 'test_notifier',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        notifier = manager.notifiers['test_notifier']
        updater = manager.updaters['test_updater']
        notifier.notify_ipv4(ipaddress.IPv4Address('1.2.3.4'))
        notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
        assert updater.published_addresses == [
            ipaddress.IPv4Address('1.2.3.4'),
            ipaddress.IPv6Network('1234::/64'),
        ]

    def test_manager_attaches_different_notifier4_and_notifier6(self):
        """Test that DDNSManager attaches two notifiers to an updater with
        different notifier4 and notifier6"""
        config = ruddr.Config(
            main={},
            notifiers={
                'test_notifier': {
                    'type': 'test',
                },
                'test_notifier2': {
                    'type': 'test',
                }
            },
            updaters={
                'test_updater': {
                    'type': 'test',
                    'notifier4': 'test_notifier',
                    'notifier6': 'test_notifier2',
                }
            },
        )
        manager = ruddr.manager.DDNSManager(config)
        notifier = manager.notifiers['test_notifier']
        notifier2 = manager.notifiers['test_notifier2']
        updater = manager.updaters['test_updater']
        notifier.notify_ipv4(ipaddress.IPv4Address('1.2.3.4'))
        notifier.notify_ipv6(ipaddress.IPv6Network('1234::/64'))
        notifier2.notify_ipv4(ipaddress.IPv4Address('5.6.7.8'))
        notifier2.notify_ipv6(ipaddress.IPv6Network('5678::/64'))
        assert updater.published_addresses == [
            ipaddress.IPv4Address('1.2.3.4'),
            ipaddress.IPv6Network('5678::/64'),
        ]

    @pytest.mark.parametrize('skipping', ['ipv4', 'ipv6', None])
    def test_notifier_not_started_when_not_attached(self, skipping):
        """Test that DDNSManager doesn't start a notifier that's not attached
        to any updater. Skipping parameter determines whether it's not attached
        because it's skipping the family it's attached to or whether it's
        simply not attached at all."""
        notifiers = {
            'test_notifier': {
                'type': 'mock',
            },
            'test_notifier2': {
                'type': 'mock',
            }
        }
        updaters = {
            'test_updater': {
                'type': 'test',
            }
        }
        if skipping == 'ipv4':
            notifiers['test_notifier']['skip_ipv4'] = 'true'
            updaters['test_updater']['notifier4'] = 'test_notifier'
            updaters['test_updater']['notifier6'] = 'test_notifier2'
        elif skipping == 'ipv6':
            notifiers['test_notifier']['skip_ipv6'] = 'true'
            updaters['test_updater']['notifier6'] = 'test_notifier'
            updaters['test_updater']['notifier4'] = 'test_notifier2'
        else:
            updaters['test_updater']['notifier'] = 'test_notifier2'

        config = ruddr.Config(main={}, notifiers=notifiers, updaters=updaters)
        manager = ruddr.manager.DDNSManager(config)
        assert len(self.notifiers) == 2

        manager.start()
        assert self.notifiers['test_notifier'].setup_count == 0
        assert self.notifiers['test_notifier2'].setup_count == 1
        assert self.notifiers['test_notifier'].check_count == 0
        assert self.notifiers['test_notifier2'].check_count == 1
        assert self.notifiers['test_notifier'].teardown_count == 0
        assert self.notifiers['test_notifier2'].teardown_count == 0

        manager.do_notify()
        assert self.notifiers['test_notifier'].setup_count == 0
        assert self.notifiers['test_notifier2'].setup_count == 1
        assert self.notifiers['test_notifier'].check_count == 0
        assert self.notifiers['test_notifier2'].check_count == 2
        assert self.notifiers['test_notifier'].teardown_count == 0
        assert self.notifiers['test_notifier2'].teardown_count == 0

        manager.stop()
        assert self.notifiers['test_notifier'].setup_count == 0
        assert self.notifiers['test_notifier2'].setup_count == 1
        assert self.notifiers['test_notifier'].check_count == 0
        assert self.notifiers['test_notifier2'].check_count == 2
        assert self.notifiers['test_notifier'].teardown_count == 0
        assert self.notifiers['test_notifier2'].teardown_count == 1

    @pytest.mark.parametrize('which', ['test_notifier', 'test_notifier2'])
    def test_notifiers_stopped_after_error(self, which):
        """Test that DDNSManager stops all notifiers if one notifier raises
        NotifierSetupError"""
        notifiers = {
            'test_notifier': {
                'type': 'mock',
            },
            'test_notifier2': {
                'type': 'mock',
            }
        }
        updaters = {
            'test_updater': {
                'type': 'test',
                'notifier4': 'test_notifier',
                'notifier6': 'test_notifier2',
            }
        }

        config = ruddr.Config(main={}, notifiers=notifiers, updaters=updaters)
        manager = ruddr.manager.DDNSManager(config)
        assert len(self.notifiers) == 2

        self.notifiers[which].setup_error = True

        with pytest.raises(ruddr.NotifierSetupError):
            manager.start()
        assert self.notifiers['test_notifier2'].stop_count == 1
        assert self.notifiers['test_notifier'].stop_count == 1



import pytest

import doubles
import ruddr.configuration


@pytest.fixture
def configfile_factory(tmp_path):
    """Fixture creating a factory for temporary config files"""
    class ConfigFileFactory:
        def __init__(self, contents):
            with open(self.filename, 'w') as f:
                for line in contents.splitlines():
                    print(line.strip(), file=f)

        @property
        def filename(self):
            return tmp_path / 'config.ini'
    return ConfigFileFactory


@pytest.fixture
def config_factory(configfile_factory):
    def factory(contents):
        configfile = configfile_factory(contents)
        config = ruddr.configuration.read_config_from_path(configfile.filename)
        config.finalize(lambda mod, typ: True, lambda mod, typ: True)
        return config
    return factory


def test_nonexistent_file(tmp_path):
    """Test opening a nonexistent path raises ConfigError"""
    with pytest.raises(ruddr.ConfigError):
        ruddr.configuration.read_config_from_path(
            tmp_path / 'nonexistent_config.ini'
        )


def test_read_config_read_error():
    """Test read error for read_config"""
    f = doubles.BrokenFile()

    with pytest.raises(ruddr.ConfigError):
        ruddr.configuration.read_config(f)


def test_read_config_from_path_read_error(tmp_path):
    """Test read error for read_config_from_path"""
    with pytest.raises(ruddr.ConfigError):
        ruddr.configuration.read_config_from_path(tmp_path / 'bad_config.ini')


def test_config_keys(config_factory):
    """Test that a configuration contains all the keys it is supposed to"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data

        [notifier.test_notifier]
        type = iface

        [updater.test_updater]
        type = standard
        notifier = test_notifier
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
    }
    assert config.notifiers == {"test_notifier": {
        "type": "iface",
        "datadir": "/var/lib/ruddr_data",
    }}
    assert config.updaters == {"test_updater": {
        "type": "standard",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier",
        "datadir": "/var/lib/ruddr_data",
    }}


def test_config_keys_custom_updater_and_notifier(config_factory):
    """Test that a configuration contains all the keys it is supposed to
    when using a custom updater and notifier (that is, the "module" key is
    present)"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data

        [notifier.test_notifier]
        type = MyNotifier
        module = notifier_module

        [updater.test_updater]
        type = MyUpdater
        module = updater_module
        notifier = test_notifier
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
    }
    assert config.notifiers == {"test_notifier": {
        "type": "MyNotifier",
        "module": "notifier_module",
        "datadir": "/var/lib/ruddr_data",
    }}
    assert config.updaters == {"test_updater": {
        "type": "MyUpdater",
        "module": "updater_module",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier",
        "datadir": "/var/lib/ruddr_data",
    }}


def test_no_global_section(config_factory):
    """Test that the configuration can be parsed when there's no ``[ruddr]``
    section and the datadir key has the correct default value"""
    config = config_factory(
        """
        [notifier.test_notifier]
        type = iface

        [updater.test_updater]
        type = standard
        notifier = test_notifier
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr",
    }
    assert config.notifiers['test_notifier']['datadir'] == "/var/lib/ruddr"
    assert config.updaters['test_updater']['datadir'] == "/var/lib/ruddr"


def test_no_updater(config_factory):
    """Test config without an updater triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface
            """
        )


def test_no_notifier(config_factory):
    """Test config without a notifier triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_extra_section_1(config_factory):
    """Test config with extra section [foo] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [foo]

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_extra_section_2(config_factory):
    """Test config with extra section [foo.] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [foo.]

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_extra_section_3(config_factory):
    """Test config with extra section [foo.bar] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [foo.bar]

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_notifier_no_name_1(config_factory):
    """Test config with [notifier] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier]
            type = iface

            [updater.test_updater]
            type = standard
            notifier =
            """
        )


def test_notifier_no_name_2(config_factory):
    """Test config with [notifier.] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.]
            type = iface

            [updater.test_updater]
            type = standard
            notifier =
            """
        )


def test_updater_no_name_1(config_factory):
    """Test config with [updater] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_updater_no_name_2(config_factory):
    """Test config with [updater.] triggers error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [updater.]
            type = standard
            notifier = test_notifier
            """
        )


def test_notifier_updater_same_name(config_factory):
    """Test notifier and updater with same name is not error"""
    config = config_factory(
        """
        [notifier.foo]
        type = iface

        [updater.foo]
        type = standard
        notifier = foo
        """
    )

    assert config.notifiers == {"foo": {
        "type": "iface",
        "datadir": "/var/lib/ruddr",
    }}
    assert config.updaters == {"foo": {
        "type": "standard",
        "notifier4": "foo",
        "notifier6": "foo",
        "datadir": "/var/lib/ruddr",
    }}


def test_two_updaters_same_name(config_factory):
    """Test two updaters with matching names is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            notifier = test_notifier

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_two_notifiers_same_name(config_factory):
    """Test two notifiers with matching names is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_notifier_without_type(config_factory):
    """Test notifier without a type is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_updater_without_type(config_factory):
    """Test updater without a type is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            notifier = test_notifier
            """
        )


def test_notifier_invalid_type_1(configfile_factory):
    """Test notifier type rejected is error"""
    configfile = configfile_factory(
        """
        [notifier.test_notifier]
        type = iface

        [updater.test_updater]
        type = standard
        notifier = test_notifier
        """
    )
    config = ruddr.configuration.read_config_from_path(configfile.filename)

    with pytest.raises(ruddr.ConfigError):
        config.finalize(
            lambda mod, typ: False,
            lambda mod, typ: True,
        )


def test_notifier_invalid_type_2(configfile_factory):
    """Test notifier module/type rejected is error"""
    configfile = configfile_factory(
        """
        [notifier.test_notifier]
        type = MyNotifier
        module = notifier_module

        [updater.test_updater]
        type = standard
        notifier = test_notifier
        """
    )
    config = ruddr.configuration.read_config_from_path(configfile.filename)

    with pytest.raises(ruddr.ConfigError):
        config.finalize(
            lambda mod, typ: False,
            lambda mod, typ: True,
        )


def test_updater_invalid_type_1(configfile_factory):
    """Test updater type rejected is error"""
    configfile = configfile_factory(
        """
        [notifier.test_notifier]
        type = iface

        [updater.test_updater]
        type = standard
        notifier = test_notifier
        """
    )
    config = ruddr.configuration.read_config_from_path(configfile.filename)

    with pytest.raises(ruddr.ConfigError):
        config.finalize(
            lambda mod, typ: True,
            lambda mod, typ: False,
        )


def test_updater_invalid_type_2(configfile_factory):
    """Test updater module/type rejected is error"""
    configfile = configfile_factory(
        """
        [notifier.test_notifier]
        type = iface

        [updater.test_updater]
        type = MyUpdater
        module = updater_module
        notifier = test_notifier
        """
    )
    config = ruddr.configuration.read_config_from_path(configfile.filename)

    with pytest.raises(ruddr.ConfigError):
        config.finalize(
            lambda mod, typ: True,
            lambda mod, typ: False,
        )


def test_duplicate_keys_global_section(config_factory):
    """Test duplicate keys in the [ruddr] section is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [ruddr]
            datadir = /var/lib/ruddr
            datadir = /var/lib/ruddr

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_duplicate_keys_notifier(config_factory):
    """Test duplicate keys in a notifier is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [ruddr]
            datadir = /var/lib/ruddr

            [notifier.test_notifier]
            type = iface
            type = iface

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            """
        )


def test_duplicate_keys_updater(config_factory):
    """Test duplicate keys in an updater is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [ruddr]
            datadir = /var/lib/ruddr

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            notifier = test_notifier
            """
        )


def test_redundant_notifier_keys(config_factory):
    """Test notifier key when notifier4 and notifier6 both present is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [ruddr]
            datadir = /var/lib/ruddr

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            notifier = test_notifier
            notifier4 = test_notifier
            notifier6 = test_notifier
            """
        )


def test_redundant_notifier_keys_global(config_factory):
    """Test notifier key when notifier4 and notifier6 both present is error
    in [ruddr]"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [ruddr]
            datadir = /var/lib/ruddr
            notifier = test_notifier
            notifier4 = test_notifier
            notifier6 = test_notifier

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            """
        )


def test_global_notifier(config_factory):
    """Test config with global notifier"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier = test_notifier

        [notifier.test_notifier]
        type = iface

        [updater.test_updater]
        type = standard
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier": "test_notifier",
    }
    assert config.notifiers == {"test_notifier": {
        "type": "iface",
        "datadir": "/var/lib/ruddr_data",
    }}
    assert config.updaters == {"test_updater": {
        "type": "standard",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier",
        "datadir": "/var/lib/ruddr_data",
    }}


def test_global_notifier46(config_factory):
    """Test config with global notifier4 and notifier6"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier4 = test_notifier
        notifier6 = test_notifier2

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [updater.test_updater]
        type = standard
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier2",
    }
    assert config.notifiers == {
        "test_notifier": {
            "type": "iface",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_notifier2": {
            "type": "iface",
            "datadir": "/var/lib/ruddr_data",
        },
    }
    assert config.updaters == {"test_updater": {
        "type": "standard",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier2",
        "datadir": "/var/lib/ruddr_data",
    }}


def test_global_and_updater_notifier_1(config_factory):
    """Test combo 1 of notifier, notifier4, notifier6, both global and local"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier4 = test_notifier
        notifier6 = test_notifier2

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [notifier.test_notifier3]
        type = iface

        [updater.test_updater]
        type = standard

        [updater.test_updater2]
        type = standard
        notifier = test_notifier3
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier2",
    }
    assert config.updaters == {
        "test_updater": {
            "type": "standard",
            "notifier4": "test_notifier",
            "notifier6": "test_notifier2",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_updater2": {
            "type": "standard",
            "notifier4": "test_notifier3",
            "notifier6": "test_notifier3",
            "datadir": "/var/lib/ruddr_data",
        },
    }


def test_global_and_updater_notifier_2(config_factory):
    """Test combo 2 of notifier, notifier4, notifier6, both global and local"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier6 = test_notifier2

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [notifier.test_notifier3]
        type = iface

        [updater.test_updater]
        type = standard

        [updater.test_updater2]
        type = standard
        notifier = test_notifier3
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier6": "test_notifier2",
    }
    assert config.updaters == {
        "test_updater": {
            "type": "standard",
            "notifier6": "test_notifier2",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_updater2": {
            "type": "standard",
            "notifier4": "test_notifier3",
            "notifier6": "test_notifier3",
            "datadir": "/var/lib/ruddr_data",
        },
    }


def test_global_and_updater_notifier_3(config_factory):
    """Test combo 3 of notifier, notifier4, notifier6, both global and local"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier6 = test_notifier2

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [notifier.test_notifier3]
        type = iface

        [updater.test_updater]
        type = standard
        notifier6 = test_notifier

        [updater.test_updater2]
        type = standard
        notifier4 = test_notifier3
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier6": "test_notifier2",
    }
    assert config.updaters == {
        "test_updater": {
            "type": "standard",
            "notifier6": "test_notifier",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_updater2": {
            "type": "standard",
            "notifier4": "test_notifier3",
            "datadir": "/var/lib/ruddr_data",
        },
    }


def test_global_and_updater_notifier_4(config_factory):
    """Test combo 4 of notifier, notifier4, notifier6, both global and local"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier = test_notifier2

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [notifier.test_notifier3]
        type = iface

        [updater.test_updater]
        type = standard
        notifier6 = test_notifier

        [updater.test_updater2]
        type = standard
        notifier4 = test_notifier3
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier": "test_notifier2",
    }
    assert config.updaters == {
        "test_updater": {
            "type": "standard",
            "notifier6": "test_notifier",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_updater2": {
            "type": "standard",
            "notifier4": "test_notifier3",
            "datadir": "/var/lib/ruddr_data",
        },
    }


def test_global_and_updater_notifier_5(config_factory):
    """Test combo 5 of notifier, notifier4, notifier6, both global and local"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier = test_notifier

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [notifier.test_notifier3]
        type = iface

        [notifier.test_notifier4]
        type = iface

        [updater.test_updater]
        type = standard
        notifier = test_notifier2

        [updater.test_updater2]
        type = standard
        notifier4 = test_notifier3
        notifier6 = test_notifier4
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier": "test_notifier",
    }
    assert config.updaters == {
        "test_updater": {
            "type": "standard",
            "notifier4": "test_notifier2",
            "notifier6": "test_notifier2",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_updater2": {
            "type": "standard",
            "notifier4": "test_notifier3",
            "notifier6": "test_notifier4",
            "datadir": "/var/lib/ruddr_data",
        },
    }


def test_global_and_updater_notifier_6(config_factory):
    """Test combo 6 of notifier, notifier4, notifier6, both global and local"""
    config = config_factory(
        """[ruddr]
        datadir = /var/lib/ruddr_data
        notifier4 = test_notifier
        notifier6 = test_notifier2

        [notifier.test_notifier]
        type = iface

        [notifier.test_notifier2]
        type = iface

        [notifier.test_notifier3]
        type = iface

        [notifier.test_notifier4]
        type = iface

        [updater.test_updater]
        type = standard

        [updater.test_updater2]
        type = standard
        notifier4 = test_notifier3
        notifier6 = test_notifier4
        """
    )

    assert config.main == {
        "datadir": "/var/lib/ruddr_data",
        "notifier4": "test_notifier",
        "notifier6": "test_notifier2",
    }
    assert config.updaters == {
        "test_updater": {
            "type": "standard",
            "notifier4": "test_notifier",
            "notifier6": "test_notifier2",
            "datadir": "/var/lib/ruddr_data",
        },
        "test_updater2": {
            "type": "standard",
            "notifier4": "test_notifier3",
            "notifier6": "test_notifier4",
            "datadir": "/var/lib/ruddr_data",
        },
    }


def test_missing_notifier(config_factory):
    """Test missing notifier, notifier4, and notifier6 is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """
            [ruddr]
            datadir = /var/lib/ruddr

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard
            """
        )


def test_missing_notifier_on_one_updater(config_factory):
    """Test missing notifier on one updater when other has it is error"""
    with pytest.raises(ruddr.ConfigError):
        config_factory(
            """[ruddr]
            datadir = /var/lib/ruddr_data

            [notifier.test_notifier]
            type = iface

            [updater.test_updater]
            type = standard

            [updater.test_updater2]
            type = standard
            notifier4 = test_notifier
            """
        )



"""Ruddr notifier that checks the IP address of a local interface on a
schedule"""

import ipaddress

from ruddr.exceptions import NotifyError, ConfigError
from ruddr.util import get_iface_addrs
from .notifier import Notifier


class IFaceNotifier(Notifier):
    """Ruddr notifier that checks the IP address of a local interface on a
    schedule"""

    def __init__(self, name, config):
        super().__init__(name, config)

        # Interface to get address from. Normally, both IPv4 and IPv6 addresses
        # are checked. If only IPv4 is needed, use "notifier4=" to attach this
        # notifier to updaters in the config instead of plain "notifier=". If
        # there are no updaters requesting IPv6 addresses from this notifier,
        # it will skip checking the IPv6 address on the interface. (Vice versa
        # for skipping IPv4.)
        try:
            self.iface = config['iface']
        except KeyError:
            self.log.critical("'iface' config option is required")
            raise ConfigError(f"{self.name} notifier requires 'iface' config "
                              "option") from None

        # IPv6 prefix: Number of bits in the network prefix. Defaults to 64,
        # but can be manually specified in case your ISP delegates a shorter
        # prefix.
        try:
            self.ipv6_prefix = int(config.get('ipv6_prefix', '64'))
        except ValueError:
            self.log.critical("'ipv6_prefix' config option must be an integer "
                              "from 1-128")
            raise ConfigError(f"'ipv6_prefix' option for {self.name} notifier "
                              "must be an integer from 1-128") from None
        if not (1 <= self.ipv6_prefix <= 128):
            self.log.critical("'ipv6_prefix' config option must be an integer "
                              "from 1-128")
            raise ConfigError(f"'ipv6_prefix' option for {self.name} notifier "
                              "must be an integer from 1-128")

        # Allow private addresses: By default, addresses in private IP space
        # (192.168.0.0/16, 10.0.0.0/8, 192.0.2.0/24, fc00::/7, 2001:db8::/32,
        # etc.) are ignored when assigned to the monitored interface. If this
        # is set to 'true', 'on', or 'yes' (case insensitive), these addresses
        # will be eligible to be picked up and sent to the notifier.
        # Non-private addresses will still take precedence, and link-local
        # addresses are always ignored.
        self.allow_private = config.get('allow_private', 'no')
        if self.allow_private.lower() in ('yes', 'true', 'on', '1'):
            self.allow_private = True
        else:
            self.allow_private = False

        self.set_check_intervals(retry_min_interval=10,
                                 retry_max_interval=600,
                                 success_interval=1800,
                                 config=config)

    def check_once(self):
        self.log.info("Checking IP addresses.")

        # Look up the interface and get the current assigned addresses
        try:
            omit_private = not self.allow_private
            ipv4s, ipv6s = get_iface_addrs(self.iface, omit_private)
        except ValueError:
            self.log.error("Interface %s does not exist", self.iface)
            raise NotifyError("Interface %s does not exist" %
                              self.iface) from None

        # None if not wanted, otherwise True if assigned, False if not assigned
        got_ipv4 = None
        got_ipv6 = None

        if self.want_ipv4():
            try:
                ipv4 = ipv4s[0]
            except IndexError:
                got_ipv4 = False
                self.log.debug("Interface %s has no IPv4 assigned",
                               self.iface)
            else:
                got_ipv4 = True
                self.notify_ipv4(ipv4)

        if self.want_ipv6():
            try:
                ipv6 = ipv6s[0]
            except IndexError:
                got_ipv6 = False
                self.log.debug("Interface %s has no IPv6 assigned",
                               self.iface)
            else:
                ipv6 = ipaddress.IPv6Interface(
                    (ipv6, self.ipv6_prefix)).network
                got_ipv6 = True
                self.notify_ipv6(ipv6)

        # Error if no wanted address was found
        if not (got_ipv4 or got_ipv6):
            self.log.warning("Interface %s has no address assigned" %
                             self.iface)
            raise NotifyError("Interface %s has no address assigned" %
                              self.iface)

        # Error for any missing wanted and needed addresses
        if self.need_ipv4() and not got_ipv4:
            self.log.warning("Interface %s has no IPv4 address assigned" %
                             self.iface)
            raise NotifyError("Interface %s has no IPv4 assigned" %
                              self.iface)
        if self.need_ipv6() and not got_ipv6:
            self.log.warning("Interface %s has no IPv6 address assigned" %
                             self.iface)
            raise NotifyError("Interface %s has no IPv6 assigned" %
                              self.iface)
